// Buffer class implementation
#include "pch.h"

#include "Buffer.h"
#include "util.h"

namespace nf::client::render {
    Buffer::Buffer(BufferType type, const VkDevice& device, VideoMemoryAllocator& allocator, const VkCommandPool& commandPool, const VkQueue& queue, void* bufferData, size_t bufferSize)
        : Resource(device)
        , m_allocator(allocator)
        , m_handle()
        , m_memory()
        , m_offset()
        , m_pointer()
    {
        if (type == BufferType::Uniform) {
            createBuffer(bufferSize, VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT, m_handle);
            m_allocator.allocateForBuffer(m_handle, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT, m_memory, m_offset);
            vkBindBufferMemory(m_device, m_handle, m_memory, m_offset);

            vkMapMemory(m_device, m_memory, m_offset, bufferSize, 0, &m_pointer);
            return;
        }

        VkBuffer stagingBuffer = nullptr;
        createBuffer(bufferSize, VK_BUFFER_USAGE_TRANSFER_SRC_BIT, stagingBuffer);

        VkDeviceMemory stagingBufferMemory = nullptr;
        VkDeviceSize stagingBufferMemoryOffset = 0;
        m_allocator.allocateForBuffer(stagingBuffer, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT, stagingBufferMemory, stagingBufferMemoryOffset);
        vkBindBufferMemory(m_device, stagingBuffer, stagingBufferMemory, stagingBufferMemoryOffset);

        void* stagingBufferPointer = nullptr;
        vkMapMemory(m_device, stagingBufferMemory, stagingBufferMemoryOffset, bufferSize, 0, &stagingBufferPointer);
        std::memcpy(stagingBufferPointer, bufferData, bufferSize);
        vkUnmapMemory(m_device, stagingBufferMemory);

        VkBufferUsageFlags mainUsage = NULL;

        switch (type) {
            case BufferType::Vertex:
                mainUsage = VK_BUFFER_USAGE_VERTEX_BUFFER_BIT;
                break;

            case BufferType::Index:
                mainUsage = VK_BUFFER_USAGE_INDEX_BUFFER_BIT;
                break;
        }

        createBuffer(bufferSize, VK_BUFFER_USAGE_TRANSFER_DST_BIT | mainUsage, m_handle);
        m_allocator.allocateForBuffer(m_handle, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT, m_memory, m_offset);
        vkBindBufferMemory(m_device, m_handle, m_memory, m_offset);

        copyBuffer(stagingBuffer, m_handle, bufferSize, commandPool, queue);

        destroyBuffer(stagingBuffer, stagingBufferMemory, stagingBufferMemoryOffset);
    }

    const VkBuffer& Buffer::getHandle() const {
        return m_handle;
    }

    void* Buffer::getPointer() const {
        return m_pointer;
    }

    void Buffer::createBuffer(VkDeviceSize size, VkBufferUsageFlags usage, VkBuffer& buffer) {
        VkBufferCreateInfo bufferCI = { VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO };
        bufferCI.size = size;
        bufferCI.usage = usage;

        if (vkCreateBuffer(m_device, &bufferCI, nullptr, &buffer) != VK_SUCCESS)
            NFError("Could not create vertex buffer.");
    }

    void Buffer::copyBuffer(VkBuffer bufferSource, VkBuffer bufferDestination, VkDeviceSize size, VkCommandPool commandPool, VkQueue queue) {
        VkCommandBuffer commandBuffer = nullptr;
        VkCommandBufferAllocateInfo commandBufferAI = { VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO };
        commandBufferAI.commandPool = commandPool;
        commandBufferAI.commandBufferCount = 1;

        if (vkAllocateCommandBuffers(m_device, &commandBufferAI, &commandBuffer) != VK_SUCCESS)
            NFError("Could not create command buffer.");

        VkCommandBufferBeginInfo commandBufferBI = { VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO };
        commandBufferBI.flags = VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT;

        if (vkBeginCommandBuffer(commandBuffer, &commandBufferBI) != VK_SUCCESS)
            NFError("Could not begin recording command buffer.");

        VkBufferCopy bufferCopyRegion = {};
        bufferCopyRegion.size = size;
        vkCmdCopyBuffer(commandBuffer, bufferSource, bufferDestination, 1, &bufferCopyRegion);

        if (vkEndCommandBuffer(commandBuffer) != VK_SUCCESS)
            NFError("Could not end command buffer.");

        VkSubmitInfo submitInfo = { VK_STRUCTURE_TYPE_SUBMIT_INFO };
        submitInfo.commandBufferCount = 1;
        submitInfo.pCommandBuffers = &commandBuffer;

        if (vkQueueSubmit(queue, 1, &submitInfo, nullptr) != VK_SUCCESS)
            NFError("Could not submit to Vulkan queue.");

        vkQueueWaitIdle(queue);

        vkFreeCommandBuffers(m_device, commandPool, 1, &commandBuffer);
    }

    void Buffer::destroyBuffer(VkBuffer buffer, VkDeviceMemory bufferMemory, VkDeviceSize offset) {
        m_allocator.deallocate(bufferMemory, offset);
        vkDestroyBuffer(m_device, buffer, nullptr);
    }

    Buffer::~Buffer() {
        destroyBuffer(m_handle, m_memory, m_offset);
    }
}
